/*******************************************************
	EEPROM
********************************************************/

#include <string.h>
#include <avr/io.h>
#include <util/twi.h>
#include "user_defs.h"

// ֐vg^Cv
void EEPROM_sequence(void);
uint8_t check_EEPROM_busy(void);
void kick_EEPROM_write(uint8_t *buf, uint16_t size);
void kick_EEPROM_read(uint16_t size);
void EEPROM_read_buffer_copy(uint8_t *buf);
static uint8_t getIIC1stCmd(uint32_t addr,uint8_t access);
static void i2c_error_stop(void);

#if 0
static uint8_t check_I2C_status(void);
#endif

uint32_t Ext_EEPROM_addr;	// EEPROMANZXAhX

static uint8_t EEPROM_sequence_number;	// EEPROMV[PXԍ

static uint8_t EEPROM_WriteBuffer[EEPROM_write_page_size];	// EEPROM݃obt@
static uint8_t EEPROM_ReadBuffer[EEPROM_write_page_size * 2];	// EEPROMǂݍ݃obt@ Rec_buffer̃TCYɍ킹Ă
static uint16_t EEPROM_WrBuf_index;	// EEPROM݃obt@ʒu
static uint16_t EEPROM_RdBuf_index;	// EEPROMǂݍ݃obt@ʒu
static uint16_t EEPROM_WriteSize;	// EEPROM݃TCY
static uint8_t EEPROM_verify;		// EEPROMׂ݂t@C{wtO

volatile uint8_t I2C_error;			// I2CُJE^

enum tag_EEPROM_sequence {
	EEPROM_IDLE = 0,
	EEPROM_WRITE_START,
	EEPROM_SET_WRITE_COMMAND,
	EEPROM_SET_WRITE_HIGH_ADDRESS,
	EEPROM_SET_WRITE_LOW_ADDRESS,
	EEPROM_SET_WRITE_DATA,
	EEPROM_WRITE_END_WAIT,
	EEPROM_READ_START,
	EEPROM_CHECK_WRITE_CYCLE,
	EEPROM_SET_READ_HIGH_ADDRESS,
	EEPROM_SET_READ_LOW_ADDRESS,
	EEPROM_SET_READ_START,
	EEPROM_SET_READ_COMMAND,
	EEPROM_SET_READ_REQUEST,
	EEPROM_GET_DATA,
	EEPROM_VERIFY
};

void SerialSendHEX_2dig(uint8_t data, uint8_t after_space);
void SerialSendHEX_4dig(uint16_t data, uint8_t after_space);
void SerialSendHEX_5dig(uint32_t data, uint8_t after_space);
void SerialSendString(const char *str);

#if 0
// eXgpi폜j
void EEPROM_test(void);
void makebuffer(uint8_t);

uint8_t EEPROM_test_step;
uint8_t EEPROM_test_data;
volatile uint32_t test_error;
uint8_t EEPROM_write_test_buffer[EEPROM_write_page_size];
uint8_t EEPROM_read_test_buffer[EEPROM_write_page_size*2];
uint8_t EEPROM_write_cycle;
volatile uint32_t adr_temp;
#endif

uint8_t check_EEPROM_busy(void){
	uint8_t result;

	result = true;
	if(EEPROM_sequence_number == EEPROM_IDLE){
		result = false;
	}
	return result;
}

const char end_of_line[] = {
	CR, LF, 0
};


void kick_EEPROM_write(uint8_t *buf, uint16_t size){
	memcpy(EEPROM_WriteBuffer,buf,size);
	EEPROM_WriteSize = size;
	EEPROM_verify = true;
	EEPROM_sequence_number = EEPROM_WRITE_START;
}

void kick_EEPROM_read(uint16_t size){
	EEPROM_WriteSize = size;
	EEPROM_verify = false;
	EEPROM_sequence_number = EEPROM_READ_START;
}

void EEPROM_read_buffer_copy(uint8_t *buf){
	memcpy(buf,EEPROM_ReadBuffer,EEPROM_WriteSize);
}
void EEPROM_sequence(void){

	switch(EEPROM_sequence_number){
		// Ȃ
		case EEPROM_IDLE:
		break;

		// ݃V[PX
		case EEPROM_WRITE_START:

#ifdef SERIAL_MONITOR_DEBUG
		SerialSendString("EEPROM WRITE ADDRESS:");
		SerialSendHEX_5dig(Ext_EEPROM_addr, false);
		SerialSendString(end_of_line);
#endif

		// X^[gRfBVs
		TWCR = _BV(TWINT) | _BV(TWSTA) | _BV(TWEN);	// X^[gRfBVs
		EEPROM_WrBuf_index = 0;
		EEPROM_sequence_number++;
		break;

		case EEPROM_SET_WRITE_COMMAND:
		if((TWCR & _BV(TWINT)) != 0){	// TWINTZbg܂ő҂
			if(TW_STATUS != TW_START){
				// G[~
				i2c_error_stop();
			}else{
				// ݃R}hs
				TWDR = getIIC1stCmd(Ext_EEPROM_addr,EEPROM_WRITE);	// o̓f[^ݒ
				TWCR = _BV(TWINT) | _BV(TWEN);	// TWINTNA }X^MJn
				EEPROM_sequence_number++;
			}
		}
		break;

		case EEPROM_SET_WRITE_HIGH_ADDRESS:
		if((TWCR & _BV(TWINT)) != 0){	// TWINTZbg܂ő҂
			if(TW_STATUS != TW_MT_SLA_ACK){
				// G[~
				i2c_error_stop();
			}else{
				// ʃAhXݒ
				TWDR = Ext_EEPROM_addr >> 8;
				TWCR = _BV(TWINT) | _BV(TWEN);	// TWINTNA }X^MJn
				EEPROM_sequence_number++;
			}
		}
		break;

		case EEPROM_SET_WRITE_LOW_ADDRESS:
		if((TWCR & _BV(TWINT)) != 0){	// TWINTZbg܂ő҂
			if(TW_STATUS != TW_MT_DATA_ACK){
				// G[~
				i2c_error_stop();
			}else{
				// ʃAhXݒ
				TWDR = Ext_EEPROM_addr;
				TWCR = _BV(TWINT) | _BV(TWEN);	// TWINTNA }X^MJn
				EEPROM_sequence_number++;
			}
		}
		break;

		case EEPROM_SET_WRITE_DATA:
		if((TWCR & _BV(TWINT)) != 0){	// TWINTZbg܂ő҂
			if(TW_STATUS != TW_MT_DATA_ACK){
				// G[~
				i2c_error_stop();
			}else{
				// f[^
				if(EEPROM_WrBuf_index < EEPROM_WriteSize){
					// weʂɒB܂ŁAЂ珑
#ifdef SERIAL_MONITOR_DEBUG
					SerialSendHEX_2dig(EEPROM_WriteBuffer[EEPROM_WrBuf_index], true);
#endif
					TWDR = EEPROM_WriteBuffer[EEPROM_WrBuf_index];
					TWCR = _BV(TWINT) | _BV(TWEN);	// TWINTNA }X^MJn
					EEPROM_WrBuf_index++;
				}else{
					// weʂ񂾂XgbvRfBV𔭍sAǂݏoV[PX
#ifdef SERIAL_MONITOR_DEBUG
					SerialSendString(end_of_line);
#endif
					TWCR = _BV(TWINT) | _BV(TWSTO) | _BV(TWEN);	// XgbvRfBVs
					EEPROM_sequence_number++;
				}
			}
		}
		break;

		case EEPROM_WRITE_END_WAIT:
		if((TWCR & _BV(TWSTO)) == 0){
			if(EEPROM_verify == true){
				EEPROM_sequence_number++;
			}else{
				EEPROM_sequence_number = EEPROM_IDLE;
			}
		}
		break;

		// ǂݏoV[PX
		case EEPROM_READ_START:	// ݑ҂
		TWCR = _BV(TWINT) | _BV(TWSTA) | _BV(TWEN);	// X^[gRfBVs
		EEPROM_RdBuf_index = 0;
		EEPROM_sequence_number++;
		break;

		case EEPROM_CHECK_WRITE_CYCLE:
		if((TWCR & _BV(TWINT)) != 0){	// TWINTZbg܂ő҂
			if((TW_STATUS != TW_START) && (TW_STATUS != TW_REP_START)){
				// G[~
				i2c_error_stop();
			}else{
				TWDR = getIIC1stCmd(Ext_EEPROM_addr,EEPROM_WRITE);	// o̓f[^ݒ
				TWCR = _BV(TWINT) | _BV(TWEN);	// TWINTNA }X^MJn
				EEPROM_sequence_number++;
			}
		}
		break;

		case EEPROM_SET_READ_HIGH_ADDRESS:
		if((TWCR & _BV(TWINT)) != 0){	// TWINTZbg܂ő҂
			if(TW_STATUS != TW_MT_SLA_ACK){
				// busygC
				TWCR = _BV(TWINT) | _BV(TWSTA) | _BV(TWEN);	// X^[gRfBVs
				EEPROM_sequence_number = EEPROM_CHECK_WRITE_CYCLE;
			}else{
				// ʃAhXݒ
				TWDR = Ext_EEPROM_addr >> 8;
				TWCR = _BV(TWINT) | _BV(TWEN);	// TWINTNA }X^MJn
				EEPROM_sequence_number++;
			}
		}
		break;

		case EEPROM_SET_READ_LOW_ADDRESS:
		if((TWCR & _BV(TWINT)) != 0){	// TWINTZbg܂ő҂
			if(TW_STATUS != TW_MT_DATA_ACK){
				// G[~
				i2c_error_stop();
			}else{
				// ʃAhXݒ
				TWDR = Ext_EEPROM_addr;
				TWCR = _BV(TWINT) | _BV(TWEN);	// TWINTNA }X^MJn
				EEPROM_sequence_number++;
			}
		}
		break;

		case EEPROM_SET_READ_START:
		if((TWCR & _BV(TWINT)) != 0){	// TWINTZbg܂ő҂
			if(TW_STATUS != TW_MT_DATA_ACK){
				// G[~
				i2c_error_stop();
			}else{
				// X^[gRfBVs
				TWCR = _BV(TWINT) | _BV(TWSTA) | _BV(TWEN);	// X^[gRfBVs
				EEPROM_sequence_number++;
			}
		}
		break;

		case EEPROM_SET_READ_COMMAND:
		if((TWCR & _BV(TWINT)) != 0){	// TWINTZbg܂ő҂
			if(TW_STATUS != TW_REP_START){
				// G[~
				i2c_error_stop();
			}else{
				// ǂݏoR}hs
				TWDR = getIIC1stCmd(Ext_EEPROM_addr,EEPROM_READ);	// o̓f[^ݒ
				TWCR = _BV(TWINT) | _BV(TWEN);	// TWINTNA }X^MJn
				EEPROM_sequence_number++;
			}
		}
		break;

		case EEPROM_SET_READ_REQUEST:
		if((TWCR & _BV(TWINT)) != 0){	// TWINTZbg܂ő҂
			if(TW_STATUS != TW_MR_SLA_ACK){
				// G[~
				i2c_error_stop();
			}else{
				// MJn
				if(EEPROM_RdBuf_index < EEPROM_WriteSize){
					if(EEPROM_RdBuf_index == (EEPROM_WriteSize - 1)){
						TWCR = _BV(TWINT) | _BV(TWEN);	// TWINTNA }X^MJn
					}else{
						TWCR = _BV(TWINT) | _BV(TWEN) | _BV(TWEA);		// TWINTNA }X^MJn(ACK)
					}
					EEPROM_sequence_number++;
				}else{
				}
			}
		}
		break;

		case EEPROM_GET_DATA:
		if((TWCR & _BV(TWINT)) != 0){	// TWINTZbg܂ő҂
			if(TW_STATUS == TW_MR_DATA_ACK){
				// f[^i[
				EEPROM_ReadBuffer[EEPROM_RdBuf_index] = TWDR;
#ifdef SERIAL_MONITOR_DEBUG
				SerialSendHEX_2dig(EEPROM_ReadBuffer[EEPROM_RdBuf_index], true);
#endif
				EEPROM_RdBuf_index++;

				// f[^ǂݏoݒ
				if(EEPROM_RdBuf_index < EEPROM_WriteSize){
					if(EEPROM_RdBuf_index == (EEPROM_WriteSize - 1)){
						TWCR = _BV(TWINT) | _BV(TWEN);	// TWINTNA }X^MJn
					}else{
						TWCR = _BV(TWINT) | _BV(TWEN) | _BV(TWEA);		// TWINTNA }X^MJn(ACK)
					}
				}
			}else if(TW_STATUS == TW_MR_DATA_NACK){
				// f[^i[
				EEPROM_ReadBuffer[EEPROM_RdBuf_index] = TWDR;
#ifdef SERIAL_MONITOR_DEBUG
				SerialSendHEX_2dig(EEPROM_ReadBuffer[EEPROM_RdBuf_index], true);
#endif
				EEPROM_RdBuf_index++;
#ifdef SERIAL_MONITOR_DEBUG
				SerialSendString(end_of_line);
#endif
				TWCR = _BV(TWINT) | _BV(TWSTO) | _BV(TWEN);	// XgbvRfBVs
				if(EEPROM_verify == true){
					EEPROM_sequence_number++;
				}else{
					EEPROM_sequence_number = 0;
				}
			}else{
				// G[~
				i2c_error_stop();
			}
		}
		break;

		case EEPROM_VERIFY:
		if((TWCR & _BV(TWSTO)) == 0){
			if(EEPROM_verify == true){
				if(memcmp(EEPROM_ReadBuffer,EEPROM_WriteBuffer,EEPROM_WriteSize) != 0){
					I2C_error++;
				}
			}
			EEPROM_sequence_number = EEPROM_IDLE;
		}
		break;

		default:
		EEPROM_sequence_number = EEPROM_IDLE;
		break;
	}
}

static void i2c_error_stop(void){
	TWCR = _BV(TWINT) | _BV(TWSTO) | _BV(TWEN);	// XgbvRfBVs
	I2C_error++;
	EEPROM_sequence_number = EEPROM_IDLE;
}

static uint8_t getIIC1stCmd(uint32_t addr,uint8_t access){
	uint8_t temp;

	if(access == 0){
		temp = 0b10100001;	// [hR}h
	}else{
		temp = 0b10100000;	// CgR}h
	}
	switch(addr >> 17){
		case 0:
		case 1:
		case 2:
			// 24FC1025p
			temp |= (addr >> 16) & 0b00000110;	// `bvZNgrbg
			if((addr & 0x10000) != 0){
				temp |= 0b00001000;	// ubNZNgrbg
			}
		break;

		case 3:
			// 24FC1026p
			temp |= (addr >> 15) & 0b00001100;	// `bvZNgrbg
			if((addr & 0x10000) != 0){
				temp |= 0b00000010;	// ubNZNgrbg
			}
		break;

		default:
			temp = 0;
		break;
	}
	return temp;
}

#if 0
static uint8_t check_I2C_status(void){
	uint8_t result;

	result = false;
	// ACKl̃`FbN
	switch(TW_STATUS) {			// TWPS1,0}XN
		case TW_MT_SLA_ACK:
		case TW_MR_SLA_ACK:
		case TW_MT_DATA_ACK:
			// ACKM
			result = true;				// ACK
		break;

		default:
		break;
	}

	return(result);
}
#endif

#if 0
void EEPROM_test(void){
	extern uint8_t EEPROM_request_flag;
	uint8_t i;

	switch(EEPROM_test_step){
		case 0:
		break;

		case 1:
//		for(i=0;i<EEPROM_write_page_size;i++){
//			EEPROM_write_test_buffer[i] = (uint8_t)(i+Ext_EEPROM_addr);
//		}
//		kick_EEPROM_write(EEPROM_write_test_buffer, EEPROM_write_page_size);
		Ext_EEPROM_addr = 0x00000;
		EEPROM_request_flag = false;
		EEPROM_write_cycle++;
		EEPROM_test_step++;
		break;

		case 2:
		if(check_EEPROM_busy() == false){
			adr_temp = (Ext_EEPROM_addr & 0xffff) + EEPROM_write_page_size;
			if(EEPROM_request_flag == false){
				if(adr_temp >= 0x10000){
					adr_temp = 0x10000 - (Ext_EEPROM_addr & 0xffff);
					makebuffer(adr_temp);
					kick_EEPROM_write(EEPROM_write_test_buffer, adr_temp);
				}else{
					makebuffer(EEPROM_write_page_size);
					kick_EEPROM_write(EEPROM_write_test_buffer, EEPROM_write_page_size);
				}
				EEPROM_request_flag = true;
			}else{
				if(adr_temp >= 0x10000){
					adr_temp = 0x10000 - (Ext_EEPROM_addr & 0xffff);
					Ext_EEPROM_addr += adr_temp;
				}else{
					Ext_EEPROM_addr += EEPROM_write_page_size;
				}
				EEPROM_request_flag = false;
				if(Ext_EEPROM_addr >= 0x80000){
//				if(Ext_EEPROM_addr >= 0x200){
					Ext_EEPROM_addr = 0;
					EEPROM_test_step++;
				}
			}
		}
		break;

		case 3:
		if(check_EEPROM_busy() == false){
			if(Ext_EEPROM_addr < 0x80000){
				if(EEPROM_request_flag == false){
					kick_EEPROM_read(EEPROM_read_page_size);
					EEPROM_request_flag = true;
				}else{
					for(i=0;i<EEPROM_read_page_size;i++){
						if(((i+Ext_EEPROM_addr) & 1) == 0){
							if(EEPROM_ReadBuffer[i] != (uint8_t)((i+Ext_EEPROM_addr) >> 8)){
								test_error++;
							}
						}else{
							if(EEPROM_ReadBuffer[i] != (uint8_t)(i+Ext_EEPROM_addr)){
								test_error++;
							}
						}
					}
					EEPROM_request_flag = false;

					Ext_EEPROM_addr += 128;
				}
			}else{
				EEPROM_test_step++;
			}
		}
		break;

		case 4:
		break;

		default:
			EEPROM_test_step = 0;
		break;
	}
}

void makebuffer(uint8_t size){
	uint8_t i;
	for(i=0;i<size;i++){
		if(((i+Ext_EEPROM_addr) & 1) == 0){
			EEPROM_write_test_buffer[i] = (i+Ext_EEPROM_addr) >> 8;
		}else{
			EEPROM_write_test_buffer[i] = (i+Ext_EEPROM_addr);
		}
	}
}
#endif
