
#include <string.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include "user_defs.h"

void GetUserConfig(void);
void SaveUserConfig(void);
void UserCommandRequest(void);
void LedDisplay(void);

static void estabish_user_setting(void);
static void estabish_beep_setting(void);
static void menu_up_down(uint8_t flag);


uint8_t progress_check(uint8_t check_progress);
void beep_set(uint8_t type);

uint8_t RS232C_speed;
uint8_t TWI_speed;
uint8_t BEEP_volume;

static uint8_t EEPROM_write_buffer[3];		// AVREEPROMǂ/obt@
static uint8_t EEPROM_check_step;			// AVREEPROM`FbNtF[Y
static uint8_t EEPROM_check_index;			// AVREEPROMe`FbNCfNX(AhX)
uint8_t menu_number;						// \j[ԍ

extern uint8_t Progress[3];					// ԑJڊǗf[^
extern volatile uint8_t key_buffer[4];		// {̃L[
extern uint8_t select_block;				// EEPROMIubNԍ
extern uint8_t PS_comminucation_error;		// SPIG[tO

static const uint16_t PROGMEM RS232C_speed_setting_tbl[]={	// RS232Cxf[^
	RS232C2X(300ul),	// 0
	RS232C2X(600ul),	// 1
	RS232C2X(1200ul),	// 2
	RS232C2X(2400ul),	// 3
	RS232C2X(4800ul),	// 4
	RS232C2X(9600ul),	// 5
	RS232C2X(19200ul),	// 6
	RS232C2X(38400ul),	// 7
	RS232C2X(57600ul),	// 8
	RS232C2X(115200ul)	// 9
};

/*******************************************************
	AVR{̋L^f[^̓ǂݏo
********************************************************/
void GetUserConfig(void){
	uint8_t i;

	for(i=0;i<ELEMENT(EEPROM_write_buffer);i++){
		EEPROM_write_buffer[i] = eeprom_read_byte((uint8_t *)((uint16_t)i));
	}

	if((EEPROM_write_buffer[0] != 0) && (EEPROM_write_buffer[0] <= ELEMENT(RS232C_speed_setting_tbl))){
		RS232C_speed = EEPROM_write_buffer[0] - 1;
		if(EEPROM_write_buffer[1] >= 7){
			TWI_speed = EEPROM_write_buffer[1];
		}else{
			TWI_speed = 7;
		}
		if((EEPROM_write_buffer[2] != 0) && (EEPROM_write_buffer[2] <= 2)){
			BEEP_volume = EEPROM_write_buffer[2] - 1;
		}else{
			BEEP_volume = 1;
		}
	}else{
		// RS232Cxf[^ُȂÃf[^ُƌߕt
		RS232C_speed = 5;
		TWI_speed = 7;
		BEEP_volume = 1;
	}

	estabish_user_setting();
}

/*******************************************************
	[U[ݒ̃n[hEFAf
********************************************************/
static void estabish_user_setting(void){
	UBRR3 = pgm_read_word(RS232C_speed_setting_tbl[RS232C_speed]);
	TWBR = TWI_speed;
	estabish_beep_setting();
}

static void estabish_beep_setting(void){
	PORTB = (PORTB & ~(_BV(PORTB4) | _BV(PORTB5))) | (BEEP_volume << PORTB4);
}
/*******************************************************
	[Uݒf[^AVREEPROMւ̏
********************************************************/
void SaveUserConfig(void){
	if(menu_number == 0){
		// ŐV̒lōXVĂ
		EEPROM_write_buffer[0] = RS232C_speed + 1;
		EEPROM_write_buffer[1] = TWI_speed;
		EEPROM_write_buffer[2] = BEEP_volume + 1;
	}else{
		// ݒ胁j[ŘMĂԂ͍XVȂ
	}

	switch(EEPROM_check_step){
	case 0:
		if(eeprom_is_ready() != 0){
			EEPROM_check_step++;
		}
	break;

	case 1:
		if(EEPROM_check_index >= ELEMENT(EEPROM_write_buffer)){
			EEPROM_check_index = 0;
		}
		if(EEPROM_write_buffer[EEPROM_check_index] == eeprom_read_byte((uint8_t *)((uint16_t)EEPROM_check_index))){
			EEPROM_check_index++;
			EEPROM_check_step = 0;
		}else{
			EEPROM_check_step++;
		}
	break;

	case 2:
		if(eeprom_is_ready() != 0){
			EEPROM_check_step++;
		}
	break;

	case 3:
		eeprom_write_byte((uint8_t *)((uint16_t)EEPROM_check_index),EEPROM_write_buffer[EEPROM_check_index]);
		EEPROM_check_index++;
		EEPROM_check_step = 0;
	break;

	default:
		EEPROM_check_index = 0;
		EEPROM_check_step = 0;
	break;
	}
}

/*******************************************************
	L[ʂɂ鑀
********************************************************/
void UserCommandRequest(void){
	if((key_buffer[trigger] & _BV(3)) != 0){
		// UPL[
		menu_up_down(true);
	}

	if((key_buffer[trigger] & _BV(2)) != 0){
		// DOWNL[
		menu_up_down(false);
	}

	if((key_buffer[trigger] & _BV(1)) != 0){
		// LEFTL[
		menu_number--;
		if(menu_number > 3){
			menu_number = 3;
		}
		if(menu_number == 0){
			estabish_user_setting();
		}
		beep_set(2);
	}

	if((key_buffer[trigger] & _BV(0)) != 0){
		// RIGHTL[
		menu_number++;
		if(menu_number > 3){
			menu_number = 0;
		}
		if(menu_number == 0){
			estabish_user_setting();
		}
		beep_set(2);
	}

	if((key_buffer[trigger] & _BV(4)) != 0){
		// RECORDL[
		if((progress_check(progress_stop) == true) && (PS_comminucation_error == 0) && (menu_number == 0)){
			Progress[target] = progress_record;
			beep_set(5);
		}else{
			beep_set(7);
		}
	}

	if((key_buffer[trigger] & _BV(5)) != 0){
		// STOPL[
		if(menu_number == 0){
			if(Progress[target] != progress_stop){
				Progress[target] = progress_stop;
				beep_set(8);
			}else{
				beep_set(7);
			}
		}else{
			menu_number = 0;
			estabish_user_setting();
			beep_set(2);
		}
	}

	if((key_buffer[trigger] & _BV(6)) != 0){
		// PLAYL[
		if((progress_check(progress_stop) == true) && (PS_comminucation_error == 0) && (menu_number == 0)){
			Progress[target] = progress_play;
			beep_set(5);
		}else{
			beep_set(7);
		}
	}

}

/*******************************************************
	j[ڃAbv/_E
********************************************************/
static void menu_up_down(uint8_t flag){
	switch(menu_number){

		case 0:
		if(progress_check(progress_stop) == true){
			// ubNԍ
			if(flag == true){
				if(select_block < 7){
					select_block++;
					beep_set(BEEP_PARAM_UP);
				}else{
					select_block = 7;
					beep_set(BEEP_COMMAND_NG);
				}
			}else{
				if(select_block > 0){
					select_block--;
					beep_set(BEEP_PARAM_DOWN);
				}else{
					select_block = 0;
					beep_set(BEEP_COMMAND_NG);
				}
			}
		}else{
			beep_set(BEEP_COMMAND_NG);
		}
		break;

		case 1:
		if(progress_check(progress_stop) == true){
			// RS232CʐMx
			if(flag == true){
				if(RS232C_speed < (ELEMENT(RS232C_speed_setting_tbl) - 1)){
					RS232C_speed++;
					beep_set(BEEP_PARAM_UP);
				}else{
					RS232C_speed = (ELEMENT(RS232C_speed_setting_tbl) - 1);
					beep_set(BEEP_COMMAND_NG);
				}
			}else{
				if(RS232C_speed > 0){
					RS232C_speed--;
					beep_set(BEEP_PARAM_DOWN);
				}else{
					RS232C_speed = 0;
					beep_set(BEEP_COMMAND_NG);
				}
			}
		}else{
			beep_set(BEEP_COMMAND_NG);
		}
		break;

		case 2:
		if(progress_check(progress_stop) == true){
			// TWIʐMx
			if(flag == true){
				if(TWI_speed < 0xff){
					TWI_speed++;
					beep_set(BEEP_PARAM_UP);
				}else{
					TWI_speed = 0xff;
					beep_set(BEEP_COMMAND_NG);
				}
			}else{
				if(TWI_speed > 7){
					TWI_speed--;
					beep_set(BEEP_PARAM_DOWN);
				}else{
					TWI_speed = 7;
					beep_set(BEEP_COMMAND_NG);
				}
			}
		}else{
			beep_set(BEEP_COMMAND_NG);
		}
		break;

		case 3:
		if(progress_check(progress_stop) == true){
			// duU[ʑ
			if(flag == true){
				if(BEEP_volume < 3){
					BEEP_volume++;
					beep_set(BEEP_PARAM_UP);
				}else{
					BEEP_volume = 3;
					beep_set(BEEP_COMMAND_NG);
				}
			}else{
				if(BEEP_volume > 0){
					BEEP_volume--;
					beep_set(BEEP_PARAM_DOWN);
				}else{
					BEEP_volume = 0;
					beep_set(BEEP_COMMAND_NG);
				}
			}
			// BEEP{[́AfĂǂB
			// AAEEPROMւ݂̏̓j[EoȂƏȂǂƎvB
			estabish_beep_setting();
		}else{
			beep_set(BEEP_COMMAND_NG);
		}
		break;

		default:
			menu_number = 0;
		break;
	}
}

/*******************************************************
	LED_/
********************************************************/
void LedDisplay(void){
	if(Progress[target] == progress_play){
		LEDB_ON();
	}else{
		LEDB_OFF();
	}
	if(Progress[target] == progress_record){
		LEDR_ON();
	}else{
		LEDR_OFF();
	}
}
