/*******************************************************
	L[L^/Đ\tg
********************************************************/

#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h> 
#include <avr/wdt.h>
#include "user_defs.h"
#include "initSFR.h"



// ֐vg^Cv
static void init(void);
void PS_sync(void);
void EEPROM_manage_read(void);
uint16_t manage_check_sum(uint8_t block);
void key_in(void);
void user_command_request(void);
void user_command_execute(void);
uint8_t progress_check(uint8_t check_progress);
void PAD_record(void);
void dump_load_proc(void);
void debug_check(uint8_t val);

// EEPROM.c
uint8_t check_EEPROM_busy(void);
void kick_EEPROM_write(uint8_t *buf, uint16_t size);
void kick_EEPROM_read(uint16_t size);
void EEPROM_read_buffer_copy(uint8_t *buf);
void EEPROM_sequence(void);
#if 0
void EEPROM_test(void);
#endif

// OLED.c
void OLED_init(void);
void OLED_sequence(void);
void OLED_display(void);

// RS232C.c
void SerialSendHEX_2dig(uint8_t data, uint8_t after_space);
void SerialSendHEX_4dig(uint16_t data, uint8_t after_space);
void SerialSendHEX_5dig(uint32_t data, uint8_t after_space);
void SerialSendString(const char *str);
void RS232C_command_check(void);
void RS232C_command_execute(void);

// UserInterface.c
void GetUserConfig(void);
void SaveUserConfig(void);
void UserCommandRequest(void);
void LedDisplay(void);

// beep.c
void beep_init(void);
void beep(void);
void beep_set(uint8_t number);
void beep_test_data_set(void);



// ϐ`
uint8_t PSCON_data_size;				// PADf[^/MTCY			
uint8_t PS_command_buffer_index;		// PADf[^Mʒu
volatile uint8_t c_same_key;			// PADf[^AJE^

uint8_t PS_data_before[32+2];			// PADf[^Ol
uint8_t PS_data_buffer[32+2];			// PADf[^l
uint8_t PS_command_buffer[32+2];		// {̎Mf[^
uint8_t PS_send_buffer[32+2];			// PADMf[^

uint8_t Rec_buffer[EEPROM_read_page_size * 2];	// L^/Đobt@

uint8_t select_block;					// EEPROMIubNԍ
uint8_t rec_manage_modify;				// EEPROMǗ̈XVtO
struct RECORD_MANAGE_INFORMATION rec_manage[8];	// EEPROMǗ

uint32_t EEPROM_RecSize;				// PADf[^L^TCY

uint16_t Rec_buffer_index;				// L^݈ʒu
uint16_t Play_buffer_index;				// Đ݈ʒu@Rec_buffer_indexœłȂH
uint8_t PS_send_buffer_index;			// Mobt@ʒu@PS_command_buffer_indexœłȂH
//uint32_t start_EEPROM_addr;			// KvȂȂ

uint8_t Progress[3];					// ԑJڊǗf[^

volatile uint8_t key_buffer[4];			// {̃L[

uint8_t EEPROM_request_flag;			// EEPROMV[PXvtO
volatile uint8_t f_stopper;				// PS-synctO
volatile uint8_t PORTD_now;				// PS-SYNCMl
volatile uint8_t PORTD_bak;				// PS-SYNCMOl

uint8_t stop_progress_step;				// STOPJڎ̏V[PXԍ

uint8_t PS_comminucation_error;			// SPIG[tO
uint8_t SYNC_recover_time;				// SYNCM҂
uint8_t PSPAD_recover_time;				// PADʐM҂
uint16_t SYNC_check;					// SYNCM`FbN(ӖC[v̗]Tx𑪂)
uint16_t SYNC_min;						// SYNCM`FbN񐔍ŏl(҂ԌŒlopr Cdlł܂܂ł͎cKv邩)

uint8_t load_data_high_nibble;			// RS232CMoCg̏ʃjutO

// EEPROM.c
extern uint32_t Ext_EEPROM_addr;		// EEPROMANZXAhX

// OLED.c
extern uint8_t VRAM_buffer[16*2];		// OLEDVRAMC[W

// RS232C.c
extern uint8_t RS232C_recv_buff[16];	// RS323CMobt@
extern uint8_t RS232C_recv_index;		// RS232CMobt@݈ʒu
extern uint8_t RS232C_read_index;		// RS232CMobt@ǂݏoʒu

volatile uint8_t EEPROM_busy_error;		// EEPROM}VXe[g҂SPĨI[oo(fobOpr̈Ӗ)

// oCi16iϊe[u
const uint8_t PROGMEM Hex2Chr[16*2] = {
	"0123456789ABCDEF"
	"0123456789abcdef"
};

const uint8_t PROGMEM Chr2Hex[256] = {
//  0    1    2    3    4    5    6    7    8    9    a    b    c    d    e    f
 0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0x10,0xff,0xff,0x11,0xff,0xff, // 00
 0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff, // 10
 0x12,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff, // 20
 0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0xff,0xff,0xff,0xff,0xff,0xff, // 30
 0xff,0x0a,0x0b,0x0c,0x0d,0x0e,0x0f,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff, // 40
 0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff, // 50
 0xff,0x0a,0x0b,0x0c,0x0d,0x0e,0x0f,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff, // 60
 0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff, // 70
 0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff, // 80
 0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff, // 90
 0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff, // a0
 0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff, // b0
 0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff, // c0
 0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff, // d0
 0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff, // e0
 0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0x13  // f0
};

extern const char end_of_line[];

volatile uint8_t debug_break;			// breakd|(fobOpr)
volatile uint8_t debug_val;				// ʒuǐ(fobOpr)

#if 1	// C[v̕׃`FbN
uint16_t main_loop_time;
uint16_t main_loop_min;
uint16_t main_loop_max;
uint32_t main_loop_ave;
#endif

/*******************************************************
	C
********************************************************/
int main(){
	cli();							// 荞݋֎~
	init();							// 낢돉
#if 1	// C[v̕׃`FbN
	main_loop_min = 0xffff;
	main_loop_max = 0x0000;
	main_loop_ave = 0xffff;
#endif
	sei();							// 荞݋

	while(1){
#if 1	// C[v̕׃`FbN
		TCNT5 = 0;
		TIFR5 |= _BV(TOV5);
#endif
		PS_sync();					// SYNCɓ鏈
#if 0
		EEPROM_test();				// EEPROMǂݏeXg
#endif
		EEPROM_sequence();			// EEPROM񓯊/ǂݏo
		RS232C_command_check();		// RS232CMR}h`FbN
		RS232C_command_execute();	// RS232CMR}hs
		dump_load_proc();			// RS232C_v/[h
		OLED_display();				// OLED\f[^
		OLED_sequence();			// OLED\]
		SaveUserConfig();			// [Uݒ`FbN/
		beep();						// BEEPo
		wdt_reset();
#if 1	// C[v̕׃`FbN
		if((TIFR5 & _BV(TOV5)) != 0){
			main_loop_time = 0xffff;
		}else{
			main_loop_time = TCNT5;
		}
		if(main_loop_min > main_loop_time){
			main_loop_min = main_loop_time;
		}
		if(main_loop_max < main_loop_time){
			main_loop_max = main_loop_time;
		}
		main_loop_ave += main_loop_time;
		main_loop_ave = main_loop_ave >> 1;
#endif

	}
}

/*******************************************************
	낢돉
********************************************************/
static void init(void){
	InitSFR();						// SFR
	GetUserConfig();				// [Uݒf[^ǂݏo
	OLED_init();					// OLED
	EEPROM_manage_read();			// EEPROMǗǂݏo

	rec_manage_modify = 0;
	Ext_EEPROM_addr = 0x40000;

	PS_command_buffer_index = 0;
	PSCON_data_size = 32;

	Progress[target] = progress_stop;
	Progress[mec] = progress_stop;
	Progress[bak] = progress_stop;

	f_stopper = true;
	PS_comminucation_error = 0;

	SYNC_min = 0xffff;
	beep_test_data_set();
	beep_init();
	wdt_reset();
}

/*******************************************************
	SYNCM
********************************************************/
void PS_sync(void){
	if(f_stopper == false){
		key_in();
		UserCommandRequest();
		user_command_execute();
		PAD_record();
		LedDisplay();
		f_stopper = true;
	}else{
		// SYNCMgKo
		PORTD_bak = PORTD_now;
		PORTD_now = PIND & _BV(PORTD2);
		SYNC_check++;
		if(PORTD_now != 0){
			if(PORTD_bak != PORTD_now){
				f_stopper = false;
				if(SYNC_recover_time == 0){
					PS_comminucation_error &= ~SYNC_ERROR;
				}else{
					SYNC_recover_time--;
				}
				if(SYNC_min > SYNC_check){
					SYNC_min = SYNC_check;
				}
				SYNC_check = 0;
				if((PS_command_buffer_index != (PSCON_data_size + 2)) || (PS_data_buffer[1] != 0x5a)){
					if(progress_check(progress_play) == false){
						PS_comminucation_error |= PSPAD_ERROR;
						PSPAD_recover_time = 20;
					}
				}else{
					if(PSPAD_recover_time == 0){
						PS_comminucation_error &= ~PSPAD_ERROR;
					}else{
						PSPAD_recover_time--;
					}
				}
				PS_command_buffer_index = 0;
			}
		}
		if(SYNC_check > 1000){
			PS_comminucation_error |= SYNC_ERROR;
			SYNC_recover_time = 10;
			f_stopper = false;
			SYNC_check = 0;
		}
	}
}

/*******************************************************
	{̃L[荞ݏ
********************************************************/
void key_in(void){
	// 荞ݒl̒ǂo
	key_buffer[before_2nd] = key_buffer[before_1st];
	key_buffer[before_1st] = key_buffer[now];

	// l̎荞/H
	key_buffer[now] = PINL & 0b00000111;
	key_buffer[now] = key_buffer[now] << 4;
	key_buffer[now] |= PINK & 0b00001111;
	key_buffer[now] ^= 0b01111111;

	// gKo
	if(key_buffer[now] == key_buffer[before_1st]){
		key_buffer[trigger] = key_buffer[before_2nd] ^ key_buffer[now];
		key_buffer[trigger] &= key_buffer[now];
	}else{
		key_buffer[trigger] = 0;
	}
}

/*******************************************************
	RS232CR}hM`FbN
********************************************************/
void RS232C_command_check(void){
	if(RS232C_recv_index != RS232C_read_index){
		if(progress_check(progress_stop) == true){
			if((RS232C_recv_buff[RS232C_read_index] == 'a') || (RS232C_recv_buff[RS232C_read_index] == 'A')){
				Ext_EEPROM_addr = 0x00000;
				EEPROM_RecSize = 0x80000;
				if((RS232C_recv_buff[(RS232C_read_index - 1) & 0x0f] == 'l') || (RS232C_recv_buff[(RS232C_read_index - 1) & 0x0f] == 'L')){
					Progress[target] = progress_load;
				}else if((RS232C_recv_buff[(RS232C_read_index - 1) & 0x0f] == 'm') || (RS232C_recv_buff[(RS232C_read_index - 1) & 0x0f] == 'M')){
					Progress[target] = progress_dump;
				}else{
				}
			}else if((RS232C_recv_buff[RS232C_read_index] >= '0') && (RS232C_recv_buff[RS232C_read_index] <= '7')){
				Ext_EEPROM_addr = (RS232C_recv_buff[RS232C_read_index] - '0') * 0x10000;
				EEPROM_RecSize = RS232C_recv_buff[RS232C_read_index] - '0' + 1;
				EEPROM_RecSize = EEPROM_RecSize << 16;
				if((RS232C_recv_buff[(RS232C_read_index - 1) & 0x0f] == 'l') || (RS232C_recv_buff[(RS232C_read_index - 1) & 0x0f] == 'L')){
					Progress[target] = progress_load;
				}else if((RS232C_recv_buff[(RS232C_read_index - 1) & 0x0f] == 'm') || (RS232C_recv_buff[(RS232C_read_index - 1) & 0x0f] == 'M')){
					Progress[target] = progress_dump;
				}else{
				}
			}else{
			}
			RS232C_read_index = (RS232C_read_index + 1) & 0x0f;
		}
	}
}

/*******************************************************
	vZX
********************************************************/
void user_command_execute(void){
	static uint8_t check_modify;
	if(Progress[target] == progress_record){
		// RECORDJ
		if(PS_comminucation_error == 0){
			if(Progress[target] != Progress[mec]){
				if(PSCON_data_size == 0){
					PSCON_data_size = 32;
				}
				memcpy(PS_data_before, &PS_data_buffer[2], PSCON_data_size/*+2*/);
				rec_manage[select_block].key_header[0] = PS_data_buffer[0];
				rec_manage[select_block].key_header[1] = PS_data_buffer[1];
				Ext_EEPROM_addr = (select_block * 0x10000) + sizeof(rec_manage[0]);
//				start_EEPROM_addr = Ext_EEPROM_addr;
				Rec_buffer_index = 0;
				rec_manage[select_block].flag &= ~CONTINUE_BIT;
//				rec_manage_modify |= _BV(select_block);
				c_same_key = 0;
				EEPROM_request_flag = false;
				Progress[mec] = Progress[target];
			}else if(Progress[bak] != Progress[mec]){
				Progress[bak] = Progress[mec];
			}else{
				// RECORDȂ̂ŉȂ
			}
		}else{
			// JڒɃG[狭~
			Progress[target] = progress_stop;
			Progress[mec] = Progress[target];
			Progress[bak] = Progress[mec];
			beep_set(9);
			debug_check(1);
		}
	}else if(Progress[target] == progress_stop){
		// STOPJ
		if(Progress[mec] == progress_record){
			stop_progress_step = 0;
			memcpy(&Rec_buffer[Rec_buffer_index], PS_data_before, PSCON_data_size);
			Rec_buffer[Rec_buffer_index + 2] = c_same_key;
			c_same_key = 0;
			kick_EEPROM_write(Rec_buffer,Rec_buffer_index+3);
			check_modify = 0;
			stop_progress_step = 0;
			Progress[mec] = Progress[target];
		}else if(Progress[mec] == progress_play){
//			memcpy(&PS_send_buffer[2],&Rec_buffer[Play_buffer_index], PSCON_data_size);
//			PS_send_buffer[0] = rec_manage[0].key_header[0];
//			PS_send_buffer[1] = rec_manage[0].key_header[1];
			PORTA = _BV(PORTA4);				// |[gAo̓WX^
			stop_progress_step = 0;
			Progress[mec] = Progress[target];
		}else if(Progress[bak] == progress_record){
			switch(stop_progress_step){
				case 0:
				if(check_EEPROM_busy() == false){
					EEPROM_RecSize = Ext_EEPROM_addr + Rec_buffer_index + PSCON_data_size + 1;
					rec_manage[select_block].rec_size = EEPROM_RecSize;
					rec_manage_modify |= _BV(select_block);
					stop_progress_step++;
				}
				break;

				case 1:
				if((rec_manage_modify & _BV(check_modify)) != 0){
					rec_manage[check_modify].sum = manage_check_sum(check_modify);
					Ext_EEPROM_addr = check_modify * 0x10000;
					kick_EEPROM_write((uint8_t *)&rec_manage[check_modify],sizeof(rec_manage[0]));
				}
				stop_progress_step++;
				break;

				case 2:
				if(check_EEPROM_busy() == false){
					rec_manage_modify &= ~_BV(check_modify);
					check_modify++;
					if(check_modify >= 8){
						Progress[bak] = Progress[mec];
					}else{
						stop_progress_step = 1;
					}
				}
				break;

				default:
				break;
			}
		}else if(Progress[bak] == progress_play){
			Progress[bak] = Progress[mec];
		}else{
			// STOPȂ̂ŉȂ
		}
	}else if(Progress[target] == progress_play){
		// PLAYJ
		if(PS_comminucation_error == 0){
			if(Progress[mec] != progress_play){
				if((rec_manage[select_block].flag & CONTINUE_BIT) != 0){
					Progress[target] = progress_stop;
					debug_check(100);
				}else if(rec_manage[select_block].rec_size == 0){
					debug_check(101);
					Progress[target] = progress_stop;
				}else if(rec_manage[select_block].rec_size > 0x80000){
					Progress[target] = progress_stop;
					debug_check(102);
				}else{
					Ext_EEPROM_addr = (select_block * 0x10000) + sizeof(rec_manage[0]);
//					start_EEPROM_addr = Ext_EEPROM_addr;
					EEPROM_RecSize = rec_manage[select_block].rec_size;
					kick_EEPROM_read(ELEMENT(Rec_buffer));
					Progress[mec] = Progress[target];
				}
			}else if(Progress[bak] != progress_play){
				if(check_EEPROM_busy() == false){
					EEPROM_read_buffer_copy(Rec_buffer);
					Ext_EEPROM_addr += ELEMENT(Rec_buffer);
					Play_buffer_index = 0;
					memcpy(&PS_send_buffer[2],&Rec_buffer[Play_buffer_index], PSCON_data_size/*+2*/);
					PS_send_buffer[0] = rec_manage[select_block].key_header[0];
					PS_send_buffer[1] = rec_manage[select_block].key_header[1];
					PORTA = _BV(PORTA3);				// |[gAo̓WX^
					c_same_key = 0;
					Progress[bak] = Progress[mec];
				}
			}else{
				// PLAYȂ̂ŉȂ
			}
		}else{
			// JڒɃG[狭~
			PORTA = _BV(PORTA4);				// |[gAo̓WX^
			Progress[target] = progress_stop;
			Progress[mec] = Progress[target];
			Progress[bak] = Progress[mec];
			beep_set(9);
			debug_check(2);
		}
	}else if(Progress[target] == progress_dump){
	}else if(Progress[target] == progress_load){
	}else{
		// ɕs΋~
		// ȌԂPLAỶ\̂ŁAMMX[ݒɂĂB
		PORTA = _BV(PORTA4);				// |[gAo̓WX^
		Progress[target] = progress_stop;
		Progress[mec] = Progress[target];
		Progress[bak] = Progress[mec];
		beep_set(9);
		debug_check(3);
	}
}

/*******************************************************
	RS232CR}hs
********************************************************/
void RS232C_command_execute(void){
	if(Progress[target] == progress_dump){
		// DUMPJ
		if(Progress[mec] != progress_dump){
//			start_EEPROM_addr = Ext_EEPROM_addr;
			kick_EEPROM_read(ELEMENT(Rec_buffer));
			Progress[mec] = Progress[target];

		}else if(Progress[bak] != progress_dump){
			if(check_EEPROM_busy() == false){
				EEPROM_read_buffer_copy(Rec_buffer);
				Ext_EEPROM_addr += ELEMENT(Rec_buffer);
				Play_buffer_index = 0;
				memcpy(&PS_send_buffer[2],&Rec_buffer[Play_buffer_index], PSCON_data_size/*+2*/);
				SerialSendString(end_of_line);
				EEPROM_request_flag = false;
				Progress[bak] = Progress[mec];
			}
		}else{
			// uȂ
		}
	}else if(Progress[target] == progress_load){
		// LOADJ
		if(Progress[mec] != progress_load){
//			start_EEPROM_addr = Ext_EEPROM_addr;
			Progress[mec] = Progress[target];

		}else if(Progress[bak] != progress_load){
			Rec_buffer_index = 0;
			load_data_high_nibble = true;
			Progress[bak] = Progress[mec];
		}else{
			// uȂ
		}
	}else{
		// uȂ
	}
}

/*******************************************************
	vZX󋵃`FbN
********************************************************/
uint8_t progress_check(uint8_t check_progress){
	uint8_t result;
	result = false;
	if(Progress[target] == check_progress){
		if(Progress[mec] == check_progress){
			if(Progress[bak] == check_progress){
				result = true;
			}else{
				// uȂ
			}
		}else{
			// uȂ
		}
	}else{
		// uȂ
	}
	return result;
}

/*******************************************************
	L[L^/Đ
********************************************************/
void PAD_record(void){
//	uint32_t adr_temp;

	if(progress_check(progress_record) == true){
		// L^tF[Y
		if(PS_comminucation_error == 0){
			if(memcmp(PS_data_before, &PS_data_buffer[2], PSCON_data_size) == 0){
				// O荞ݓeƓȂJEgAbv
				c_same_key++;
				if(c_same_key == 0){
					// JE^ꂽobt@fo
					memcpy(&Rec_buffer[Rec_buffer_index], &PS_data_before[0], PSCON_data_size);
					Rec_buffer_index += PSCON_data_size;
					Rec_buffer[Rec_buffer_index] = 0xff;
					Rec_buffer_index++;
				}else{
					// uȂ
				}
			}else{
				// O荞ݓeƈقȂ΁A܂ł̋L^fo
				memcpy(&Rec_buffer[Rec_buffer_index], PS_data_before, PSCON_data_size);
				Rec_buffer_index += PSCON_data_size;
				Rec_buffer[Rec_buffer_index] = c_same_key;
				Rec_buffer_index++;
				c_same_key = 0;
			}
			memcpy(PS_data_before, &PS_data_buffer[2], PSCON_data_size);
			if(Rec_buffer_index >= EEPROM_write_page_size){
				// L^obt@y[WTCY𒴂Ă珑݃V[PXN
				if(check_EEPROM_busy() == false){
					if(Ext_EEPROM_addr < (0x80000 - PSCON_data_size - 1)){
						if(EEPROM_request_flag == false){
							kick_EEPROM_write(Rec_buffer, EEPROM_write_page_size);
							EEPROM_request_flag = true;
						}else{
							memcpy(Rec_buffer,&Rec_buffer[EEPROM_write_page_size],EEPROM_write_page_size);
							Rec_buffer_index -= EEPROM_write_page_size;
							Ext_EEPROM_addr += EEPROM_write_page_size;
							EEPROM_request_flag = false;
							if((Ext_EEPROM_addr & 0xffff) == 0){
								Ext_EEPROM_addr += EEPROM_write_page_size;
								rec_manage[Ext_EEPROM_addr >> 16].flag &= ~CONTINUE_BIT;
								rec_manage_modify |= _BV(Ext_EEPROM_addr >> 16);
							}else{
								// uȂ
							}
						}
					}else{
						Progress[target] = progress_stop;
						beep_set(8);
						debug_check(4);
					}
				}else{
					// uȂ
				}
			}else{
				// uȂ
			}

			if(Rec_buffer_index >= (EEPROM_write_page_size + (EEPROM_write_page_size / 2))){
				// L^obt@75%ȓŏ݂IĂė~
				// EEPROM݃V[PXԂɍĂȂIII
				EEPROM_busy_error++;
			}
		}else{
			// L^ɒʐMG[~
			Progress[target] = progress_stop;
			beep_set(9);
			debug_check(5);
		}
	}else if(progress_check(progress_play) == true){
		// ĐtF[Y
		if(PS_comminucation_error == 0){
			if((Ext_EEPROM_addr - ELEMENT(Rec_buffer) + Play_buffer_index) < EEPROM_RecSize){
				memcpy(&PS_send_buffer[2], &Rec_buffer[Play_buffer_index], PSCON_data_size);
//				PS_send_buffer[0] = rec_manage[select_block].key_header[0];
//				PS_send_buffer[1] = rec_manage[select_block].key_header[1];
				if(c_same_key >= Rec_buffer[Play_buffer_index+PSCON_data_size]){
					c_same_key = 0;
					Play_buffer_index += (PSCON_data_size + 1);
				}else{
					c_same_key++;
				}
				if(Play_buffer_index >= EEPROM_read_page_size){
					// Đobt@ʒuy[WTCY𒴂Ă玟y[WǂݏoNGXg
					if(check_EEPROM_busy() == false){
						if(EEPROM_request_flag == false){
							kick_EEPROM_read(EEPROM_read_page_size);
							EEPROM_request_flag = true;
						}else{
							memcpy(Rec_buffer,&Rec_buffer[EEPROM_write_page_size],EEPROM_read_page_size);
							EEPROM_read_buffer_copy(&Rec_buffer[EEPROM_write_page_size]);
							Play_buffer_index -= EEPROM_write_page_size;
							Ext_EEPROM_addr += EEPROM_write_page_size;
							if((Ext_EEPROM_addr & 0xffff) == 0){
								Ext_EEPROM_addr += EEPROM_write_page_size;
							}else{
								// uȂ
							}
							EEPROM_request_flag = false;
						}
					}else{
						// uȂ
					}
				}else{
					// uȂ
				}
			}else{
				Progress[target] = progress_stop;
				PORTA = _BV(PORTA4);				// |[gAo̓WX^
				beep_set(8);
				debug_check(6);
			}
		}else{
			// ĐɒʐMG[~
			Progress[target] = progress_stop;
			PORTA = _BV(PORTA4);				// |[gAo̓WX^
			beep_set(9);
			debug_check(7);
		}
	}else if(progress_check(progress_dump) == true){
	}else if(progress_check(progress_load) == true){
	}else{
		// uȂ
	}
}

/*******************************************************
	RS232C(_v/[h)
********************************************************/
void dump_load_proc(void){
	uint8_t temp_data;
	uint8_t temp_buffer_index;

	if(progress_check(progress_dump) == true){
		// _vtF[Y
		if((Ext_EEPROM_addr - ELEMENT(Rec_buffer) + Play_buffer_index) < EEPROM_RecSize){
			if(Play_buffer_index < ELEMENT(Rec_buffer)){	// Đobt@鎖͖ƎvÔ
				SerialSendHEX_2dig(Rec_buffer[Play_buffer_index], true);
				Play_buffer_index++;
				if((Play_buffer_index & 0x000f) == 0){
					SerialSendString(end_of_line);
				}else{
					// uȂ
				}
			}
			if(Play_buffer_index >= EEPROM_read_page_size){
				// Đobt@ʒuy[WTCY𒴂Ă玟y[WǂݏoNGXg
				if(check_EEPROM_busy() == false){
					if(EEPROM_request_flag == false){
						kick_EEPROM_read(EEPROM_read_page_size);
						EEPROM_request_flag = true;
					}else{
						memcpy(Rec_buffer,&Rec_buffer[EEPROM_write_page_size],EEPROM_read_page_size);
						EEPROM_read_buffer_copy(&Rec_buffer[EEPROM_write_page_size]);
						if((Ext_EEPROM_addr & 0xffff) == 7*0x80){
							debug_break++;
						}
						Play_buffer_index -= EEPROM_write_page_size;
						Ext_EEPROM_addr += EEPROM_write_page_size;
						EEPROM_request_flag = false;
					}
				}else{
					// uȂ
				}
			}else{
				// uȂ
			}
		}else{
			Progress[target] = progress_stop;
			Progress[mec] = Progress[target];
			Progress[bak] = Progress[mec];
			debug_check(8);
		}
	}else if(progress_check(progress_load) == true){
		temp_buffer_index = RS232C_recv_index;
		while(temp_buffer_index != RS232C_read_index){
			temp_data = pgm_read_byte(&Chr2Hex[RS232C_recv_buff[RS232C_read_index]]);
			if(temp_data <= 0x0f){
				if(load_data_high_nibble == true){
					load_data_high_nibble = false;
					Rec_buffer[Rec_buffer_index] = temp_data << 4;
				}else{
					Rec_buffer[Rec_buffer_index] |= temp_data;
					Rec_buffer_index++;
					load_data_high_nibble = true;
				}
			}else if(temp_data == 0x13){
				Progress[target] = progress_stop;
				debug_check(9);
			}else if(temp_data == 0x10){
				load_data_high_nibble = true;
			}else if(temp_data == 0x11){
				load_data_high_nibble = true;
			}else if(temp_data == 0x12){
				load_data_high_nibble = true;
			}else{
				Progress[target] = progress_stop;
				Progress[mec] = Progress[target];
				Progress[bak] = Progress[mec];
				debug_check(10);
			}
			RS232C_read_index = (RS232C_read_index + 1) & 0x0f;
		}
		if(Rec_buffer_index >= EEPROM_write_page_size){
			// L^obt@y[WTCY𒴂Ă珑݃V[PXN
			if(check_EEPROM_busy() == false){
				if((Ext_EEPROM_addr & 0xffff) == 7*0x80){
					debug_break++;
				}else{
					// uȂ
				}
				if(Ext_EEPROM_addr < 0x80000){
					if(EEPROM_request_flag == false){
						kick_EEPROM_write(Rec_buffer, EEPROM_write_page_size);
						EEPROM_request_flag = true;
					}else{
						memcpy(Rec_buffer,&Rec_buffer[EEPROM_write_page_size],EEPROM_write_page_size);
						Rec_buffer_index -= EEPROM_write_page_size;
						Ext_EEPROM_addr += EEPROM_write_page_size;
						EEPROM_request_flag = false;
					}
				}else{
					Progress[target] = progress_stop;
					Progress[mec] = Progress[target];
					Progress[bak] = Progress[mec];
					debug_check(11);
				}
			}else{
				// uȂ
			}
		}else if(Ext_EEPROM_addr >= EEPROM_RecSize){
			Progress[target] = progress_stop;
			Progress[mec] = Progress[target];
			Progress[bak] = Progress[mec];
			debug_check(12);
		}else{
			// uȂ
		}

		if(Rec_buffer_index >= (EEPROM_write_page_size + (EEPROM_write_page_size))){
			// L^obt@ꂽG[~悤ǂ悤BBB
			// EEPROM݃V[PXԂɍĂȂ˂BBB
			EEPROM_busy_error++;
			debug_check(200);
		}
	}else if(progress_check(progress_record) == true){
	}else if(progress_check(progress_play) == true){
	}else if(progress_check(progress_stop) == true){
	}else{
	}
}


/*******************************************************
	EEPROMǗ̈ǂݏo
********************************************************/
void EEPROM_manage_read(void){
	uint8_t i;

	for(i=0;i<8;i++){
		Ext_EEPROM_addr = i * 0x10000;
		kick_EEPROM_read(EEPROM_read_page_size);
		while(check_EEPROM_busy() == true){
			// 狭IɃ[vEo郋[gȁ`
			EEPROM_sequence();
			// ŃEHb`hbOɈ悤ȂATWIXs[hxI
		}
		EEPROM_read_buffer_copy((uint8_t *)&rec_manage[i]);
		if(manage_check_sum(i) != rec_manage[i].sum){
			memset(&rec_manage[i], 0, sizeof(rec_manage[0]));
		}
		wdt_reset();	// TWI̐ݒɂẮAEHb`hbOɈ\B
	}
}

/*******************************************************
	EEPROMǗf[^`FbNTZ
********************************************************/
uint16_t manage_check_sum(uint8_t block){
	uint8_t i;
	uint16_t sum;
	uint8_t *adr;

	sum = 0;
	adr = (uint8_t *)&rec_manage[block];
	for(i = 0;i < (sizeof(rec_manage[0]) - 2);i++){
		sum += (uint16_t)(*adr);
		adr++;
	}
	return sum;
}

/*******************************************************
	PADf[^M荞
********************************************************/
ISR(SPI_STC_vect){
	register uint8_t PSBASE_dat;
	register uint8_t PSCON_dat;

	PSBASE_dat = SPDR;
	PSCON_dat = PINC;

	if(PSBASE_dat == 0x01){
		if(progress_check(progress_play) == true){
			SPDR = PS_send_buffer[0];
			PS_send_buffer_index = 1;
		}else{
			// f[^t[Jn
			PS_command_buffer_index = 0;
			PS_send_buffer_index = 0;
			memset(PS_command_buffer, 0, ELEMENT(PS_command_buffer));
			memset(PS_data_buffer, 0, ELEMENT(PS_data_buffer));
		}
	}else{
		if(progress_check(progress_play) == true){
			SPDR = PS_send_buffer[PS_send_buffer_index];
			PS_send_buffer_index++;
			if(PS_send_buffer_index >= (PSCON_data_size + 3)){
				// f[^t[I
				PS_send_buffer_index = 0;
			}else{
				// uȂ
			}
		}else{
			if(PS_command_buffer_index >= (PSCON_data_size + 2)){	// obt@͈̓`FbN
//				PS_command_buffer_index = 0;
			}else{
				// uȂ
			}
			// VAf[^ۑ
			PS_command_buffer[PS_command_buffer_index] = PSBASE_dat;
			PS_data_buffer[PS_command_buffer_index] = PSCON_dat;
			// f[^t[Zo
			if(PS_command_buffer_index == 0){
				PSCON_data_size = (PSCON_dat & 0x0f) << 1;	// 4rbgt[([hf[^Z)
				if(PSCON_data_size == 0){
					PSCON_data_size = 32;	// 0͍̎ő咷ł鎖Ă
				}else{
					// uȂ
				}
			}else{
				// uȂ
			}

			// f[^t[I[`FbN
			if(PS_command_buffer_index < (PSCON_data_size + 2)){
				// f[^t[I
				PS_command_buffer_index++;
			}else{
				// uȂ
			}
		}
	}
}

void debug_check(uint8_t val){
	debug_val = val;
}


